/*  Copyright (c) 2007-2008 Tomasz Moń <desowin@gmail.com>
 *  Copyright (c) 2004 Artur Gajda <ag@net13.org>
 *  Copyright (c) 2004 Jakub Zawadzki <darkjames@darkjames.ath.cx>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; under version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses>.
 */

#include <gtk/gtk.h>
#include <libintl.h>
#include <pwd.h>
#include <stdlib.h>
#include "riv.h"
#include "conf.h"
#include "net.h"
#include <gtk/gtk.h>

gchar *plik_conf;
GrivConfig conf;

GrivConfig default_config = {
    NULL,         /* nick            */
    NULL,         /* user            */
    "ANY",        /* iface           */
    0,            /* gender          */
    0,            /* text color R    */
    255,          /* text color G    */
    0,            /* text color B    */
    0,            /* use colors      */
    S_PORT,       /* port            */
    0,            /* start minimised */
    120,          /* away timeout    */
};

void set_default() {
    memcpy(&conf, &default_config, sizeof(GrivConfig));
    struct passwd *pwd_entry = getpwuid(getuid());
    conf.user = g_strndup(pwd_entry->pw_name, 31);
    conf.nick = g_strndup(pwd_entry->pw_name, 29);
}

typedef struct conf_string_t {
    char const *name;
    gchar **item;
    int max;           /* maximum length of string; 0 for unlimited */
} conf_string;

static conf_string conf_strings[] = {
    {"nick", &conf.nick, 29},
    {"user", &conf.user, 31},
    {"iface", &conf.iface, 0},
};

static gint nconf_strings = G_N_ELEMENTS(conf_strings);

typedef struct conf_integer_t {
    const char *name;
    gint *item;
} conf_integer;

static conf_integer conf_integers[] = {
    {"gender", &conf.gender},
    {"textR", &conf.color_r},
    {"textG", &conf.color_g},
    {"textB", &conf.color_b},
    {"display_colors", &conf.display_colors},
    {"port", &conf.port},
    {"run_minimised", &conf.run_minimised},
    {"away_timeout", &conf.away_timeout},
};

static gint nconf_integers = G_N_ELEMENTS(conf_integers);

void save_config() {
    FILE *f =  fopen(plik_conf, "w");
    if (!f) {
        printf(_("Settings cannot be saved. Please check write permissions to %s\n"), plik_conf);
        return;
    }

    int x;
    for (x=0; x<nconf_strings; ++x)
         fprintf(f,"%s=%s\n", conf_strings[x].name, *conf_strings[x].item);

    for (x=0; x<nconf_integers; ++x)
         fprintf(f,"%s=%d\n", conf_integers[x].name, *conf_integers[x].item);

    fclose(f);
    printf(_("Settings saved.\n"));
}

void parse_config () {
    int i;
    set_default();
    plik_conf = g_strdup_printf("%s/.config/griv.conf", g_get_home_dir());

    if(!g_file_test(plik_conf, G_FILE_TEST_EXISTS)) {
        printf(_("Loaded default settings.\n"));
        return;
    }
    gchar* file_content;
    gchar** values;
    gchar** item;
    int x;

    g_file_get_contents(plik_conf, &file_content, NULL, NULL);
    values = g_strsplit(file_content, "\n", 0);

    for (x=0; x<nconf_strings; ++x) {
         for (i=0; values[i] && strlen(values[i]); i++) {
              item = g_strsplit(values[i], "=", 2);
              if (item[0] && item[1]) {
                  if (!strcmp(item[0], conf_strings[x].name)) {
                      if (conf_strings[x].max)
                          *conf_strings[x].item = g_strndup(item[1], conf_strings[x].max);
                      else
                          *conf_strings[x].item = g_strdup(item[1]);
                  }
              }
              g_strfreev(item);
         }
    }

    for (x=0; x<nconf_integers; ++x) {
         for (i=0; values[i] && strlen(values[i]); i++) {
              item = g_strsplit(values[i], "=", 2);
              if (item[0] && item[1]) {
                  if (!strcmp(item[0], conf_integers[x].name)) {
                          *conf_integers[x].item = atoi(item[1]);
                  }
              }
              g_strfreev(item);
         }
    }

    /* sanity checks */
    conf.color_r = MIN(MAX(0, conf.color_r), 255);
    conf.color_g = MIN(MAX(0, conf.color_g), 255);
    conf.color_b = MIN(MAX(0, conf.color_b), 255);

    if (conf.gender != 0 && conf.gender != 1 && conf.gender != 2)
        conf.gender = 0;

    if (conf.away_timeout < 0)
        conf.away_timeout = 120;

    g_strfreev(values);
    g_free(file_content);
}
