/*  Copyright (c) 2007-2008 Tomasz Moń <desowin@gmail.com>
 *  Copyright (c) 2004 Artur Gajda
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; under version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses>.
 */

#include <gtk/gtk.h>
#include <libintl.h>
#include "riv.h"
#include "lista.h"
#include "conf.h"
#include "ui_priv.h"
#include "ui.h"

#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <ifaddrs.h>
#include <string.h>

int moj_id = 0;

GList* get_my_ip() {
    GList *list = NULL;
    struct ifaddrs *ifa = NULL, *ifp = NULL;

    if (getifaddrs (&ifp) < 0) {  // can't get addresses
        return NULL;
    }

    char ip[200];
    ip[0] = '\0';
    for (ifa = ifp; ifa; ifa = ifa->ifa_next) {
         socklen_t salen;

         if (ifa->ifa_addr->sa_family == AF_INET)
             salen = sizeof(struct sockaddr_in);
         else
             continue;

         if (getnameinfo(ifa->ifa_addr, salen, ip, sizeof(ip), NULL, 0, NI_NUMERICHOST) < 0)
             continue;

#ifdef DEBUG
         printf("Found address: %s\n", ip);
#endif
         list = g_list_append(list, g_strdup(ip));
    }
    freeifaddrs(ifp);

    return list;
}

int gen_moj_id() {
/*
    There was comment before I forked griv that
    ids are generated by ip, and example was 192.168.0.26 = 436250816
    and commenter said he don't know how it's counted
    basing on the example id I regognised it as:
    436250816 = 0x1A00A8C0
    0xC0 = 192
    0xA8 = 168
    0x00 = 0
    0x1A = 26
    But RivChat clients in my network seems to have id calculated in different way

    Hence the code to generate id is commented
                                                --desowin
*/

/*
    GList *ip = get_my_ip();
    if (ip) {
        int id = inet_addr(g_list_last(ip)->data);
        g_list_free(ip);
        if (id != -1)
            return id;
    }
*/

    return time(NULL);
}

void lista_init() {
    struct INFO info;
    int i;
    for (i = 0; i < MAXLUDZI; i++) {
        lista[i].empty = TRUE;
        lista[i].away = FALSE;
        lista[i].composing = FALSE;
        lista[i].id = -1;
    }

    moj_id = gen_moj_id();

    GList *ip = get_my_ip();
    lista_update(g_get_host_name(), conf.nick, ip ? g_list_last(ip)->data : "\0", moj_id, &info);
    g_list_free(ip);
}

int lista_update(const char *host, const char *nick, const char *ip, int id, struct INFO *info) {
    int i;

    #ifdef DEBUG
    printf("lista update %s %s %s %i\n", host, nick, ip, id);
    #endif

    //sprawdzanie czy juz jest ten id
    for (i = 0; i < MAXLUDZI; i++) {
        if (lista[i].empty) continue; //skip it
        if (lista[i].id == id) {
            PRIV *priv = priv_lista_find_by_nick(lista[i].nick);
            if (priv && (priv->id != id)) priv->id = id;

            gboolean refresh_list = FALSE;

            lista[i].last_ping = time(NULL);
            if (lista[i].away != info->away) {
                lista[i].away = info->away;
                refresh_list = TRUE;
            }

            if (lista[i].composing != info->pisze) {
                lista[i].composing = info->pisze;
                refresh_list = TRUE;
            }

            memcpy(&lista[i].last_rcinfo, info, sizeof(struct INFO));

            if (refresh_list)
                gtk_widget_queue_draw(tree);
            return 0; //update tylko
        }
    }
    //dodajemy
    for (i = 0; i < MAXLUDZI; i++) {
        if (lista[i].empty) {
            #ifdef DEBUG
            printf("lista update dodalem %s %s\n", host, nick);
            #endif
            strcpy(lista[i].nick, nick);
            strcpy(lista[i].host, host);
            strcpy(lista[i].ip, ip);

            lista[i].last_ping = time(NULL);
            lista[i].empty = FALSE;
            lista[i].id = id;
            memcpy(&lista[i].last_rcinfo, info, sizeof(struct INFO));

            PRIV *priv = priv_lista_find_by_nick(nick);
            if (priv) {
                gchar *txt = g_strdup_printf("%s - %s", "griv", nick);
                gtk_window_set_title(GTK_WINDOW(priv->window), txt);
                g_free(txt);
                txt = g_strdup_printf(_("%s is online.\n"), nick);
                priv_add_text(txt, priv);
                #ifdef DEBUG
                printf("%s has changed id from %i to %i\n", nick, priv->id, id);
                #endif
                priv->id = id;
                g_free(txt);
            }

            return 1; //zmiany na liscie
        }
    }

    return -1; //nie powinno nas tu byc
}

int lista_change_nick(int id, char *newnick) {
    int i;
    for (i = 0; i < MAXLUDZI; i++) {
        if (lista[i].empty) continue;

        if (lista[i].id == id) {
            PRIV *priv = priv_lista_find_by_nick(lista[i].nick);
            gchar *nick = g_strdup(lista[i].nick);
            strcpy(lista[i].nick, newnick);
            if (priv) {
                gchar *txt = g_strdup_printf("%s - %s", "griv", newnick);
                gtk_window_set_title(GTK_WINDOW(priv->window), txt);
                g_free(txt);
                txt = g_strdup_printf(_("%s is changing nick. His new nick is %s.\n"), nick, newnick);
                priv_add_text(txt, priv);
                g_free(txt);
                g_free(priv->nick);
                priv->nick = g_strdup(newnick);
            }
            g_free(nick);

            return 1;
        }
    }
    return -1;
}

int lista_del_id(int id) {
    int i;
    //printf("lista wywalam: %s\n", host);
    for (i = 0; i < MAXLUDZI; i++) {
         if (lista[i].empty) continue;
         if (lista[i].id == id) {
             lista[i].empty = TRUE;
             PRIV *priv = priv_lista_find_by_id(id);
             if(priv) priv_set_zombie(priv, ZOMBIE_TIMEOUT);
             lista[i].id = -1;
             return 0;
         }
    }
    return -1;
}

int lista_get_gid_by_id(gint id) {
    int i;
    for (i = 0; i < MAXLUDZI; i++) {
        if (lista[i].empty) continue;
        if (lista[i].id == id) {
            return i;
        }
    }
    return -1;
}

gchar* lista_get_ip_by_id(gint id) {
    int i;
    for (i = 0; i < MAXLUDZI; i++) {
        if (lista[i].empty) continue;
        if (lista[i].id == id) {
            return lista[i].ip;
        }
    }
    return NULL;
}

gboolean lista_check_nick_availability(gchar *nick) {
    int i;
    for (i = 0; i < MAXLUDZI; i++) {
        if (lista[i].empty) continue;
        if (strcmp(lista[i].nick, nick) == 0) {
            return FALSE;
        }
    }
    return TRUE;

}

int lista_set_away(gint id, gboolean away, gchar *away_txt) {
    int i;

    for (i = 0; i < MAXLUDZI; i++) {
        if (lista[i].empty) continue;
        if (lista[i].id == id) {

            if (lista[i].away == away)
                return -1; //already marked

            #ifdef DEBUG
            if (away == 1)
                printf("%s went away \"%s\"\n", lista[i].nick, away_txt);
            else
                printf("%s is back\n", lista[i].nick);
            #endif

            lista[i].away = away;

            if (away == 1 && away_txt)
                strcpy(lista[i].away_txt, away_txt);
            else
                lista[i].away_txt[0] = '\0';

            gtk_widget_queue_draw(tree);

            return 0;
        }
    }
    return -1;
}
