/*  Copyright (c) 2007-2008 Tomasz Moń <desowin@gmail.com>
 *  Copyright (c) 2004 Artur Gajda <ag@net13.org>
 *  Copyright (c) 2004 Jakub Zawadzki <darkjames@darkjames.ath.cx>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; under version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses>.
 */

#include <stdio.h>
#include <libintl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>

#include "net.h"
#include "riv.h"
#include "ui.h"
#include "conf.h"
#include "lista.h"
#include "stuff.h"
#include "ui_priv.h"
#include "autoaway.h"

int nr = 0;
gchar *topic = NULL;
gboolean away = FALSE;
gboolean ping_away = FALSE;
gchar *away_msg = NULL;
int ssock, rsock;
struct sockaddr_in toaddr;
int slowa;
int online = 0; /* program running time, 1 = 10s. */
gboolean composing = FALSE; /* user types message */
#ifndef HAVE_SCRNSAVER
gboolean autoaway = FALSE;
#endif

void set_rcping(struct INFO *info) {
    info->away = (away || autoaway);
    info->master = 0;
    gchar *user = utf2win(conf.user);
    strcpy(info->user, user);
    g_free(user);
    info->kod = 0;
    info->gender = conf.gender;
    info->filetransfer = 0;
    info->pisze = composing;
    info->online = online;
    info->slowa = slowa;
    strncpy(info->host, g_get_host_name(), 49);
    strncpy(info->os, os_name, 19);
    strncpy(info->prog, VER, 17);
    /*
       the way RivChat numbers version doesn't match griv release numbering,
       so we're sending "\0\0" as version number.
       RivChat seems to use first byte for major version number and
       second byte for minor version number.
       so 2.35 is 0x02 0x23
                                --desowin
    */
    strcpy(info->version, "\0\0");
}

void net_setdefault(void) {
    gchar *gnick = utf2win(conf.nick);
    strcpy(hsnd.nick, gnick);
    g_free(gnick);
    memcpy(hsnd.header, "RivChat\0\0\0\0\0", 11);
    hsnd.format[0] = conf.color_r;    //red
    hsnd.format[1] = conf.color_g;    //green
    hsnd.format[2] = conf.color_b;    //blue
    hsnd.format[3] = nr;
    hsnd.format[4] = conf.gender;
    hsnd.format[5] = 0;
    hsnd.format[6] = 0;
    hsnd.size = sizeof(struct PAKIET);
    hsnd.toid = -1;
    hsnd.fromid = moj_id;
}

int sendpacket(struct PAKIET* pakiet, int typ, int toid, char *datagram) {
    pakiet->type = typ;
    if ((typ  == RC_PING) || (typ == RC_INIT))
        set_rcping(&hsnd.data.rc_ping);
    else {
        if (datagram) {
            if (strlen(datagram) >= 255) {
                strncpy(pakiet->data.data, datagram, 254);
                pakiet->data.data[255] = '\0';
            } else
                strcpy(pakiet->data.data, datagram);
        } else
            pakiet->data.data[0] = '\0';
    }

    pakiet->size = sizeof(struct PAKIET);
    pakiet->format[3] = nr++; //id pakietu ma byc rozne
    pakiet->toid = toid;

    if (toid != -1) {
        in_addr_t ip = inet_addr(lista_get_ip_by_id(toid)); /* send message to user ip */
        /* sending private message to other RivChat client running
           on localhost results in 'undelivered' message for some reason.
           Send such messages over broadcast as it seems to work then */
        if (ip != -1 &&
            ip != inet_addr(lista_get_ip_by_id(moj_id))) {
            struct sockaddr_in privaddr;
            privaddr.sin_addr.s_addr = ip;
            privaddr.sin_port = htons(conf.port); /* port number */
            privaddr.sin_family = AF_INET;
            return sendto(ssock, pakiet, sizeof(struct PAKIET), 0, (struct sockaddr *) &privaddr, sizeof(privaddr));
        } else { /* send it on broadcast */
            printf("sending private message over broadcast\n");
        }
    }

    return sendto(ssock, pakiet, sizeof(struct PAKIET), 0, (struct sockaddr *) &toaddr, sizeof(toaddr));
}


void ping_watek() {
    int now;
    int old = time(NULL) + 5;
    while (1) {
        now = time(NULL);
        if ( old + 9 < now ) {
            online++;
            old = now;
            if (ssock > -1)
                send_ping();
            int i;
            for (i = 0; i < MAXLUDZI; i++) {
                if (lista[i].empty) continue;
                if (lista[i].last_ping + PING_TIMEOUT < now) {
                    gchar *txt = g_strdup_printf(_("! %s (%s): ping timeout\n"), lista[i].nick, lista[i].host);
                    add_text(txt);
                    g_free(txt);
                    GDK_THREADS_ENTER();
                    lista_del_id(lista[i].id);
                    lista_refresh();
                    GDK_THREADS_LEAVE();
                }
            }
        }
        sleep(10);
    }
}

void recv_watek() {
    gchar *gmsg, *gnick;
    gchar *txt;
    while (1) {
        struct sockaddr_in oth;
        socklen_t sinlen = sizeof(struct sockaddr);

        struct PAKIET pkt;
        ssize_t msglen = recvfrom(rsock, &pkt, sizeof(pkt), 0, (struct sockaddr *)&oth, &sinlen);

        if (msglen != -1) {
            char ip[16];
            sprintf(ip, "%s", inet_ntoa(oth.sin_addr));
#ifdef DEBUG
            printf("recvpacket() size %zd from %s(%s) typ: %i toid: %i fromid: %i idpkt: %i\n", msglen, pkt.nick, ip, pkt.type, pkt.toid, pkt.fromid, pkt.format[3]);
#endif

            GDK_THREADS_ENTER();
            gnick = win2utf(pkt.nick);
            switch (pkt.type) {
                case (RC_MESSAGE):
                     gmsg = win2utf(pkt.data.data);

                     gchar *tag = NULL;
                     if (conf.display_colors) {
                         update_user_text_color(gnick, pkt.format[0], pkt.format[1], pkt.format[2]);
                         tag = g_strdup_printf("color-%s", gnick);
                     }

                     if (pkt.toid == moj_id)
                         priv_new_message(gnick, pkt.fromid, gmsg, tag);
                     else if (pkt.toid == -1)
                         add_text2(gnick, gmsg, tag);

                     if (tag) g_free(tag);
                     g_free(gmsg);
                     break;
            case (RC_POP):
                if (pkt.fromid == moj_id) break;
                gmsg = win2utf(pkt.data.data);
                pop_show(pkt.fromid, gnick, gmsg);
                g_free(gmsg);
                break;
            case (RC_REPOP):
                gmsg = g_strdup_printf(_("%s read your pop message\n"), gnick);
                add_info_text(gmsg);
                g_free(gmsg);
                break;
            case (RC_ME):
                gmsg = win2utf(pkt.data.data);
                txt = g_strdup_printf("* %s %s\n", gnick, gmsg);
                add_text3("FG_NAVY", txt);
                g_free(gmsg);
                g_free(txt);
                break;
            case (RC_INIT):
                gmsg = win2utf(pkt.data.rc_ping.host);
                lista_update(gmsg, gnick, ip, pkt.fromid, &pkt.data.rc_ping);
                g_free(gmsg);
                lista_refresh();

                if (pkt.fromid == moj_id) break;
                if (topic && strlen(topic) > 0)
                    send_topic(topic);
                send_ping();
                txt = g_strdup_printf(_("%s from %s (%s) has connected.\n"), gnick, pkt.data.rc_ping.host, ip);
                add_info_text(txt);
                g_free(txt);
                break;
            case (RC_PING) :
                if (lista_update(pkt.data.rc_ping.host, gnick, ip, pkt.fromid, &pkt.data.rc_ping) == 1 )
                    lista_refresh();
                break;
            case (RC_QUIT) :
                lista_del_id(pkt.fromid);
                lista_refresh();
                if (pkt.fromid == moj_id) break;
                txt = g_strdup_printf(_("%s has left.\n"), gnick);
                add_info_text(txt);
                g_free(txt);
                break;
            case (RC_TOPIC):
                gmsg = win2utf(pkt.data.data);
                if (topic) g_free(topic);
                topic = g_strdup(gmsg);
                txt = g_strdup_printf("griv: %s", gmsg);
                gtk_window_set_title(GTK_WINDOW(window), txt);
                g_free(gmsg);
                g_free(txt);
                break;
            case (RC_NEWTOPIC):
                gmsg = win2utf(pkt.data.data);
                if (topic) g_free(topic);
                topic = g_strdup_printf("%s (%s)", gmsg, gnick);
                txt = g_strdup_printf(_("%s is changing topic. The new topic is: \"%s\".\n"), gnick, gmsg);
                add_info_text(txt);
                g_free(txt);
                txt = g_strdup_printf("griv: %s", topic);
                gtk_window_set_title(GTK_WINDOW(window), txt);
                g_free(txt);
                g_free(gmsg);
                break;
            case (RC_PINGAWAY):
                gmsg = win2utf(pkt.data.data);
                lista_set_away(pkt.fromid, TRUE, gmsg);
                gint id = lista_get_gid_by_id(pkt.fromid);
                if (id != -1)
                    lista[id].last_ping = time(NULL);
                g_free(gmsg);
                break;
            case (RC_AWAY):
                gmsg = win2utf(pkt.data.data);
                if (lista_set_away(pkt.fromid, TRUE, gmsg) == -1) break; 
                txt = g_strdup_printf(_("%s is away: \"%s\".\n"), gnick, gmsg);
                add_info_text(txt);
                g_free(gmsg);
                g_free(txt);
                break;
            case (RC_REAWAY):
                if (lista_set_away(pkt.fromid, FALSE, NULL) == -1)  break;
                txt = g_strdup_printf(_("%s is coming back.\n"), gnick);
                add_info_text(txt);
                g_free(txt);
                break;
            case (RC_NICKCHANGE):
                gmsg = win2utf(pkt.data.data);
                txt = g_strdup_printf(_("%s is changing nick. His new nick is %s.\n"), gnick, gmsg);
                if (pkt.fromid == moj_id)
                    strncpy(hsnd.nick, pkt.data.data, 30);
                lista_change_nick(pkt.fromid, gmsg);
                lista_refresh();
                add_info_text(txt);
                g_free(gmsg);
                g_free(txt);
                break;
            }
            g_free(gnick);
            GDK_THREADS_LEAVE();
        }
    }
}

void send_ping() {
    if ((away || autoaway) && ping_away) {
        gchar *txt = NULL;
        if (!away)
            txt = utf2win(_("User is away"));
        sendpacket(&hsnd, RC_PINGAWAY, -1, away ? away_msg : txt);
        if (txt) g_free(txt);
    } else
        sendpacket(&hsnd, RC_PING, -1, NULL);

    if (away || autoaway)
        ping_away = !ping_away;
}

int send_message(char *txt) {
   gchar *gmsg = utf2win(txt);
   if (gmsg == NULL) return -1;
   return sendpacket(&hsnd, RC_MESSAGE, -1, gmsg);
}

int send_priv_message(char *txt, int toid) {
   gchar *gmsg = utf2win(txt);
   if (gmsg == NULL) return -1;
   return sendpacket(&hsnd, RC_MESSAGE, toid, gmsg);
}

int send_pop_message(const char *txt, int toid) {
   gchar *gmsg = utf2win(txt);
   if (gmsg == NULL) return -1;
   return sendpacket(&hsnd, RC_POP, toid, gmsg);
}

int send_me(char *txt) {
   gchar *gmsg = utf2win(txt);
   if (gmsg == NULL) return -1;
   return sendpacket(&hsnd, RC_ME, -1, gmsg);
}

int send_new_topic(char *topic) {
   gchar *gtopic = utf2win(topic);
   if (gtopic == NULL) return -1;
   return sendpacket(&hsnd, RC_NEWTOPIC, -1, gtopic);
}

int send_topic(char *topic) {
   gchar *gtopic = utf2win(topic);
   if (gtopic == NULL) return -1;
   return sendpacket(&hsnd, RC_TOPIC, -1, gtopic);
}

int send_away(char *txt) {
   gchar *gtxt = utf2win(txt);
   if (gtxt == NULL) return -1;
   away = TRUE;
   ping_away = FALSE;
   if (away_msg)
       g_free(away_msg);
   away_msg = g_strdup(gtxt);
   return sendpacket(&hsnd, RC_AWAY, -1, gtxt);
}

int send_reaway() {
   away = FALSE;
   return sendpacket(&hsnd, RC_REAWAY, -1, NULL);
}

void send_quit() {
   sendpacket(&hsnd, RC_QUIT, -1, "Papa");
}

void change_nick(gchar *nick) {
    if (!lista_check_nick_availability(nick)) {
        gchar *txt = g_strdup_printf(_("Nick \"%s\" is already taken\n"), nick);
        add_info_text(txt);
        g_free(txt);
        return;
    }

    if (conf.nick) g_free(conf.nick);
    conf.nick = g_strndup(nick, 29);
    gchar *nnick = utf2win(conf.nick);
    sendpacket(&hsnd, RC_NICKCHANGE, -1, nnick);
    g_free(nnick);
}

int net_start() {
    net_setdefault();

#ifdef DEBUG
    printf("Starting net thread...\n");
#endif

    ssock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    rsock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);

    if (ssock == -1) printf("Unable to create send socket\n");
    #ifdef DEBUG
    else             printf("Send socket: OK\n");
    #endif

    if (rsock  == -1) printf("Unable to create recv socket\n");
    #ifdef DEBUF
    else              printf("Recv socket: OK\n");
    #endif

    int buf_size = 1;
    setsockopt(rsock, SOL_SOCKET, SO_BROADCAST, &buf_size, sizeof(buf_size) ); // to ma byc przed bind().
    setsockopt(ssock, SOL_SOCKET, SO_BROADCAST, &buf_size, sizeof(buf_size) ); // to ma byc przed bind(). potrzebne (?)
    setsockopt(rsock, SOL_SOCKET, SO_REUSEADDR, &buf_size, sizeof(buf_size)); // dzieki temu nawet jesli jakis port korzysta z udp 0.0.0.0:S_PORT to sie zbinduje.
    setsockopt(ssock, SOL_SOCKET, SO_REUSEADDR, &buf_size, sizeof(buf_size));

    buf_size = BUF;
    setsockopt(rsock, SOL_SOCKET, SO_RCVBUF, &buf_size, sizeof(buf_size));
    setsockopt(ssock, SOL_SOCKET, SO_SNDBUF, &buf_size, sizeof(buf_size));

    struct sockaddr_in outaddr;

    outaddr.sin_family = AF_INET;
    outaddr.sin_port = htons( conf.port );
    if ( strcmp(conf.iface, "ANY") == 0 )
    outaddr.sin_addr.s_addr = htonl(INADDR_ANY);
    else
    outaddr.sin_addr.s_addr = inet_addr(conf.iface);

    if ( outaddr.sin_addr.s_addr == INADDR_NONE )
	    printf("Bad address\n");

    if ((bind(ssock, (struct sockaddr *) &outaddr, sizeof(struct sockaddr)) < 0) ||
        (bind(rsock, (struct sockaddr *) &outaddr, sizeof(struct sockaddr)) < 0)) {

        printf("Port %d is already used.", conf.port);
        return -1;
    }
    #ifdef DEBUG
    else printf("Bind : OK\n");
    #endif

    toaddr.sin_addr.s_addr = INADDR_BROADCAST; /* send message to 255.255.255.255 */
    toaddr.sin_port = htons( conf.port ); /* port number */
    toaddr.sin_family = AF_INET;

    if ( sendpacket(&hsnd, RC_INIT, -1, NULL) == -1 )
    	printf(_("Error whilst sending RC_INIT, program won't work\n"));

    GError *error = NULL;
    g_thread_create((GThreadFunc)recv_watek, NULL, FALSE, &error);
    if (error) {
        printf("Error creating recv thread: %s", error->message);
        g_error_free(error);
        error = NULL;
    }

    g_thread_create((GThreadFunc)ping_watek, NULL, FALSE, &error);
    if (error) {
        printf("Error creating ping thread: %s", error->message);
        g_error_free(error);
        error = NULL;
    }

    return 0;
}
